/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.cluster;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.Future;
import java.util.concurrent.atomic.AtomicInteger;

import org.apache.log4j.Logger;
import org.jgroups.Address;
import org.jgroups.ChannelException;
import org.jgroups.JChannel;
import org.jgroups.Message;
import org.jgroups.ReceiverAdapter;
import org.jgroups.View;
import org.jgroups.util.Streamable;
import org.jgroups.util.Util;

import com.koutra.dist.proc.cluster.Request.Type;
import com.koutra.dist.proc.util.ExecutorsHelper;

/**
 * A node in the cluster of cooperating processes. Each node receives and executes
 * <code>IClusterRunnable</code> instances from the cluster executor service.
 * 
 * @author Pafsanias Ftakas
 */
public class ClusterNode {
	private static final Logger logger = Logger.getLogger(ClusterNode.class);
	
	protected JChannel channel;
	protected AtomicInteger rank;
	protected AtomicInteger clusterSize;
	protected TaskMonitoringThread taskMonitoringThread;

	/**
	 * Initializing constructor. Connects to the cluster.
	 * @throws ChannelException
	 */
	public ClusterNode() throws ChannelException {
		rank = new AtomicInteger(0);
		clusterSize = new AtomicInteger(0);
		taskMonitoringThread = new TaskMonitoringThread();
		taskMonitoringThread.start();
		channel = new JChannel("udp.xml");
		channel.setReceiver(new ReceiverAdapter() {
			@Override
			public void receive(Message msg) {
				Request req = null;
				try {
					req = (Request) Util.objectFromByteBuffer(msg.getBuffer());
				} catch (Exception e) {
					logger.error("Unable to convert received message to request", e);
					return;
				}
				if (logger.isTraceEnabled())
					logger.trace("Cluster node received message " + req.getType() +
							" (" + req.getTaskID() + ")");
				
				switch (req.getType()) {
				case EXECUTE:
					logger.trace("Received execute request.");
					Entry taskEntry = new Entry(req.getTaskID(), req.getPayload(),
							msg.getSrc(), null);
					if (req.getTaskID().getId() % clusterSize.get() == rank.get()) {
						logger.trace("I need to execute this task.");
						ClusterRunnable runnable = new ClusterRunnable(req.getPayload());
						runnable.setConainingObject(req.getPayload());
						Future<?> future = ExecutorsHelper.getInstance().submitInProc(
								runnable);
						taskEntry.setFuture(future);
						taskMonitoringThread.addEntry(taskEntry);
					}
					break;
				case RESULT:
					logger.trace("Ignoring result response.");
					break;
				default:
					logger.trace("Unknown request type. Ignoring.");
				}
			}
			
			@Override
			public void viewAccepted(View new_view) {
				logger.info("Cluster view has changed: " + new_view);
				Address localAddress = channel.getLocalAddress();
				Vector<Address> addresses = new_view.getMembers();
				int new_rank = -1;
				for (int i = 0; i < addresses.size(); i++) {
					if (localAddress.equals(addresses.get(i))) {
						new_rank = i;
						break;
					}
				}
				if (new_rank == -1)
					throw new RuntimeException("We were unable to find our node in " +
							"the view. Something is seriously wrong.");
				rank.set(new_rank);
				clusterSize.set(new_view.size());
				
				if (logger.isTraceEnabled()) {
					logger.trace("Cluster size is now: " + new_view.size() +
							", my rank is: " + new_rank);
				}
			}
		});
		channel.connect("DistProcCluster");
	}
	
	public void shutdown() {
		channel.disconnect();
		channel.close();
	}
	
	private class TaskMonitoringThread extends Thread {
		private Set<Entry> entries;
		
		public TaskMonitoringThread() {
			super("TaskMonitoring");
			entries = Collections.synchronizedSet(new HashSet<Entry>());
		}
		
		public void addEntry(Entry entry) {
			entries.add(entry);
		}
		
		@Override
		public void run() {
			while (true) {
				try { Thread.sleep(500); } catch (InterruptedException ie) { }
				
				Set<Entry> entriesToRemove = new HashSet<Entry>();
				synchronized (entries) {
					for (Entry entry : entries) {
						if (entry.getFuture().isDone()) {
							try {
								Request req = new Request(
										Type.RESULT, null, entry.getTaskID(),
										(Streamable) entry.getFuture().get());
								Message msg = new Message(
										entry.getOriginator(), channel.getLocalAddress(),
										Util.objectToByteBuffer(req));
								channel.send(msg);
							} catch (Exception e) {
								logger.error("Unable to send response for result (" +
										entry.getTaskID() + ")", e);
							}
							entriesToRemove.add(entry);
						}
					}
				}
				
				entries.removeAll(entriesToRemove);
			}
		}
	}
	
	/**
	 * Start up a cluster node in this JVM.
	 * 
	 * @param args command line arguments.
	 */
	public static void main(String[] args) {
		try {
			new ClusterNode();
			while (true) {
				try {
					Thread.sleep(60 * 1000);
				} catch (InterruptedException ie) {
					// ignore
				}
			}
		} catch (Exception e) {
			logger.error("Unable to run cluster node", e);
		}
	}
}
