/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.cluster;

import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.Socket;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.util.ByteArrayUtil;

/**
 * Used in conjunction with a <code>PipedWriter</code>, this class is an reader
 * that can read the character stream from a socket, thereby allowing pipelines
 * to perform part of their computation on a remote JVM.
 * 
 * @author Pafsanias Ftakas
 */
public class PipedReader extends Reader implements IPipedInput {
	private static final Logger logger = Logger.getLogger(PipedReader.class);

	protected InetSocketAddress socketAddress;
	protected Socket socket;
	protected String charset;
	protected Reader reader;

	/**
	 * Initializing constructor. Starts a socket on the local machine and attempts to
	 * connect to the server socket specified by the <code>remoteHost</code> and
	 * <code>remotePort</code> parameters.
	 * @param remoteHost the remote server host address
	 * @param remotePort the remote server port number
	 * @throws IOException iff something goes wrong while trying to create a socket
	 * and connect to the remote server.
	 */
	public PipedReader(InetAddress remoteHost, int remotePort,
			String charset) throws IOException {
		this.socketAddress = new InetSocketAddress(InetAddress.getLocalHost(), 0);
		this.socket = new Socket();
		this.socket.bind(socketAddress);
		this.socket.connect(new InetSocketAddress(remoteHost, remotePort));
		this.charset = charset;
		this.reader = new InputStreamReader(socket.getInputStream(), charset);
	}
	
	/**
	 * This helper method is used to communicate the server details of the final
	 * BoundaryTemplate back to the originating pipeline for use in the mux instance.
	 * @param host the IP address of the host.
	 * @param port the port of the server socket.
	 * @throws IOException iff something goes awry.
	 */
	public void communicateOtherEndDetails(byte[] host, int port) throws IOException {
		String details = ByteArrayUtil.encodeInHex(host, 0, host.length) + "|" +
				String.valueOf(port) + "|\n";
		
		if (logger.isTraceEnabled())
			logger.trace("Sending details: '" + details.trim() + "' to the other end.");
		
		socket.getOutputStream().write(details.getBytes());
		// DO NOT close the reader here. It will close the damn socket!
	}

	@Override
	public int read(char[] cbuf, int off, int len) throws IOException {
		return reader.read(cbuf, off, len);
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

}
