/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.faucet;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collections;
import java.util.List;

import javax.sql.DataSource;

import org.apache.log4j.Logger;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.util.DataSourceManager;

/**
 * A faucet implementation that queries a database and uses the resulting
 * java.sql.ResultSet to feed the transformation process.
 * 
 * @author Pafsanias Ftakas
 */
public class ResultSetFaucet implements IFaucet {
	private static final Logger logger = Logger.getLogger(ResultSetFaucet.class);
	
	protected String id;
	protected boolean hookedUp;
	protected ISink sink;
	protected DataSource ds;
	protected String dataSourceName;
	protected String query;
	protected ResultSet resultSet;
	protected Statement statement;
	protected Connection connection;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public ResultSetFaucet() { }
	
	public ResultSetFaucet(String id, String dataSourceName, String query) {
		this.id = id;
		this.hookedUp = false;
		this.sink = null;
		this.ds = null;
		this.dataSourceName = dataSourceName;
		this.query = query;
		this.resultSet = null;
		this.statement = null;
		this.connection = null;
	}

	public ResultSetFaucet(String id, DataSource ds, String query) {
		this.id = id;
		this.hookedUp = false;
		this.sink = null;
		this.ds = ds;
		this.dataSourceName = null;
		this.query = query;
		this.resultSet = null;
		this.statement = null;
		this.connection = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case ResultSet:
			return true;
		case ByteStream:
		case CharStream:
		case XML:
		default:
			return false;
		}
	}

	@Override
	public void dispose() {
		try {
			if (resultSet != null)
				resultSet.close();
		} catch (SQLException e) {
			throw new XformationException("Unable to close result set");
		}
		try {
			if (statement != null)
				statement.close();
		} catch (SQLException e) {
			throw new XformationException("Unable to close statement");
		}
		try {
			if (connection != null)
				connection.close();
		} catch (SQLException e) {
			throw new XformationException("Unable to close connection");
		}
	}

	@Override
	public void hookupSink(ISink sink) {
		if (!sink.supportsInput(ContentType.ResultSet))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the ResultSet content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
		if (hookedUp && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUp)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.sink = sink;
		this.hookedUp = true;
		
		this.sink.hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case CharStream:
		case XML:
			return null;
		}
		
		if (resultSet != null)
			return resultSet;

		try {
			if (dataSourceName != null) {
				ds = DataSourceManager.getInstance().getDataSource(dataSourceName);
			}
			if (logger.isTraceEnabled())
				logger.trace("Executing query: " + query);
			connection = ds.getConnection();
			connection.setAutoCommit(true);
			statement = connection.createStatement(
					ResultSet.TYPE_FORWARD_ONLY, ResultSet.CONCUR_READ_ONLY);
			statement.setFetchSize(Integer.MIN_VALUE);
			resultSet = statement.executeQuery(query);
		} catch (SQLException e) {
			throw new XformationException("Unable to query database", e);
		}
		
		return resultSet;
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUp = true;
		this.sink.hookupFaucet(this);
		this.ds = null;
		this.dataSourceName = in.readUTF();
		this.query = in.readUTF();
		this.resultSet = null;
		this.statement = null;
		this.connection = null;
	}

	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (dataSourceName == null)
			throw new XformationException(
					"Only ResultSetFaucet instances that have the data source name " +
					"specified can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
		out.writeUTF(dataSourceName);
		out.writeUTF(query);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
