/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Future;

import org.apache.log4j.Logger;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;
import com.koutra.dist.proc.xml.IStreamSAXCallback;
import com.koutra.dist.proc.xml.StreamSAXEventReader;

public abstract class StreamToXMLPipelineItem implements ISimplePipelineItem,
		IStreamSAXCallback {
	private static final Logger logger = Logger.getLogger(StreamToXMLPipelineItem.class);
	
	protected String id;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected ISink sink;
	protected IFaucet faucet;
	protected XMLReader reader;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public StreamToXMLPipelineItem() { }
	
	public StreamToXMLPipelineItem(String id) {
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = null;
		this.faucet = null;
		this.reader = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
			return true;
		case CharStream:
		case XML:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.ByteStream))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the ByteStream content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}

	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.XML))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the XML content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpSink)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.sink = sink;
		this.hookedUpSink = true;
		
		this.sink.hookupFaucet(this);
	}

	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case CharStream:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
			" is not supported.");
		}
		
		if (reader != null)
			return reader;
		
		InputStream inputStream = (InputStream) faucet.getSource(ContentType.ByteStream);
		reader = new StreamSAXEventReader(inputStream, this);
		
		if (logger.isTraceEnabled())
			logger.trace("Creating SAX event reader " + reader +
					" mapping input stream: " + inputStream);
		
		return reader;
	}
	
	@Override
	public void registerSource(Object source) {
		// only used in mux/demux cases.
	}
	
	@Override
	public void handleMuxSwitch() {
		// only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// only used in mux/demux cases.
		
	}
	
	@Override
	public String dumpPipeline() {
		return sink.dumpPipeline() + "\n" + getClass().getName() + ": " + reader;
	}
	
	@Override
	public void dispose() {
		faucet.dispose();
	}
	
	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}
	
	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: faucet has not been set");
		
	    if (faucet instanceof IPipelineItem)
	    	((IPipelineItem) faucet).consume(this);
	    
	    if (this instanceof ISinkTemplate) {
	    	if (!((ISinkTemplate) this).isClone()) {
	    		return null;
	    	}
	    }

		if (!hookedUpSink)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: sink has not been set");

	    getSource(ContentType.XML);

		return new InputSource();
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUpSink = true;
		this.sink.hookupFaucet(this);
		this.faucet = null;
		this.hookedUpFaucet = false;
		this.reader = null;
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
