/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.pipeline;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Future;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.stream.StreamSource;

import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;

import com.koutra.dist.proc.model.ContentType;
import com.koutra.dist.proc.model.IDemuxFaucet;
import com.koutra.dist.proc.model.IFaucet;
import com.koutra.dist.proc.model.IFaucetTemplate;
import com.koutra.dist.proc.model.IMuxSink;
import com.koutra.dist.proc.model.IPipelineItem;
import com.koutra.dist.proc.model.ISimplePipelineItem;
import com.koutra.dist.proc.model.ISink;
import com.koutra.dist.proc.model.ISinkTemplate;
import com.koutra.dist.proc.model.ISplitPipelineItem;
import com.koutra.dist.proc.model.XformationException;

/**
 * A pipeline item that can be used as part of a XML transformation chain. When
 * using the constructor that takes an input stream to the style sheet, the caller
 * is responsible for closing the input stream once the transformation has completed.
 * 
 * @author Pafsanias Ftakas
 */
public class XMLFilterPipelineItem implements ISimplePipelineItem {
	private enum Type {
		File,
		InputStream;
	};
	
	protected String id;
	protected boolean hookedUpFaucet;
	protected boolean hookedUpSink;
	protected ISink sink;
	protected IFaucet faucet;
	protected Type type;
	protected Object faucetObject;
	protected InputStream styleSheetInputStream;
	protected String styleSheetPath;
	protected XMLFilter filter;
	
	/**
	 * @deprecated Use any of the initializing constructors instead.
	 */
	public XMLFilterPipelineItem() { }
	
	public XMLFilterPipelineItem(String id, InputStream styleSheetInputStream) {
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = null;
		this.faucet = null;
		this.type = Type.InputStream;
		this.faucetObject = null;
		this.styleSheetInputStream = styleSheetInputStream;
		this.styleSheetPath = null;
		this.filter = null;
	}

	public XMLFilterPipelineItem(String id, String styleSheetPath) {
		this.id = id;
		this.hookedUpFaucet = false;
		this.hookedUpSink = false;
		this.sink = null;
		this.faucet = null;
		this.type = Type.File;
		this.faucetObject = null;
		this.styleSheetInputStream = null;
		this.styleSheetPath = styleSheetPath;
		this.filter = null;
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public ISink getSink() {
		return sink;
	}

	@Override
	public IFaucet getFaucet() {
		return faucet;
	}

	@Override
	public boolean supportsInput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}

	@Override
	public boolean supportsOutput(ContentType contentType) {
		switch (contentType) {
		case XML:
			return true;
		case ByteStream:
		case CharStream:
		case ResultSet:
		default:
			return false;
		}
	}
	
	protected void checkFaucetValidity(IFaucet faucet) {
		if (!faucet.supportsOutput(ContentType.XML))
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' must support the XML content type.");
		if (faucet instanceof IDemuxFaucet)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IDemuxFaucet instance. This is not allowed.");
		if (faucet instanceof IFaucetTemplate)
			throw new IllegalArgumentException("Faucet '" + faucet.getId() +
					"' is an IFaucetTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupFaucet(IFaucet faucet) {
		checkFaucetValidity(faucet);
		if (hookedUpFaucet && this.faucet == faucet) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpFaucet)
			throw new XformationException("Trying to hook up an already hooked up sink.");
		
		this.faucet = faucet;
		this.hookedUpFaucet = true;
		
		this.faucet.hookupSink(this);
	}
	
	protected void checkSinkValidity(ISink sink) {
		if (!sink.supportsInput(ContentType.XML))
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' must support the XML content type.");
		if (sink instanceof IMuxSink)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an IMuxSink instance. This is not allowed.");
		if (sink instanceof ISinkTemplate)
			throw new IllegalArgumentException("Sink '" + sink.getId() +
					"' is an ISinkTemplate instance. This is not allowed.");
	}

	@Override
	public void hookupSink(ISink sink) {
		checkSinkValidity(sink);
		if (hookedUpSink && this.sink == sink) {
			// Effectively a NO-OP.
			return;
		}
		if (hookedUpSink)
			throw new XformationException("Trying to hook up an already hooked up faucet.");
		
		this.sink = sink;
		this.hookedUpSink = true;
		
		this.sink.hookupFaucet(this);
	}
	
	@Override
	public Object getSource(ContentType contentType) {
		switch (contentType) {
		case ByteStream:
		case CharStream:
		case ResultSet:
			throw new XformationException("Content type: " + contentType +
					" is not supported.");
		}

		if (filter != null)
			return filter;
		
		switch (type) {
		case File:
			try {
				styleSheetInputStream = new BufferedInputStream(
						new FileInputStream(styleSheetPath));
			} catch (FileNotFoundException e) {
				throw new XformationException("Unable to create input stream", e);
			}
			break;
		case InputStream:
			break;
		}
		try {
			SAXTransformerFactory tf = (SAXTransformerFactory)
					TransformerFactory.newInstance();
			filter = tf.newXMLFilter(new StreamSource(styleSheetInputStream));
		} catch (TransformerConfigurationException e) {
			throw new XformationException("Unable to create filter", e);
		}

		return filter;
	}
	
	@Override
	public void registerSource(Object source) {
		// Only used in mux/demux cases.
	}
	
	@Override
	public void handleMuxSwitch() {
		// Only used in mux/demux cases.
	}
	
	@Override
	public void endMux() {
		// Only used in mux/demux cases.
		
	}
	
	@Override
	public String dumpPipeline() {
		return sink.dumpPipeline() + "\n" + getClass().getName() + ": " + filter;
	}
	
	@Override
	public void dispose() {
		faucet.dispose();
	}
	
	@Override
	public void consume() {
		throw new XformationException("Calling ISink.consume() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Future<?> consumeAsynchronously() {
		throw new XformationException("Calling ISink.consumeAsynchronously() on a IPipelineItem " +
				"implementation. Call IPipelineItem.consume(ISink) instead.");
	}

	@Override
	public Object consume(ISink previousSink) {
		if (!hookedUpFaucet)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: faucet has not been set");
		
		Object inputSource = null;
		if (faucet instanceof IPipelineItem)
			inputSource = ((IPipelineItem) faucet).consume(this);
		
	    if (this instanceof ISinkTemplate) {
	    	if (!((ISinkTemplate) this).isClone()) {
	    		return null;
	    	}
	    }

		if (!hookedUpSink)
			throw new XformationException("Pipeline item has not been set up " +
					"correctly: sink has not been set");

		faucetObject = faucet.getSource(ContentType.XML);
		
		getSource(ContentType.XML);
		
		if (faucetObject instanceof XMLReader) {
			XMLReader parent = (XMLReader) faucetObject;
			filter.setParent(parent);
			
		    if (faucet instanceof ISink)
		    	return inputSource;
		    else
		    	return null;// TODO should we throw an exception here?
		} else {
		    try {
				SAXParserFactory spf = SAXParserFactory.newInstance();
			    spf.setNamespaceAware(true);
			    SAXParser parser = spf.newSAXParser();
			    XMLReader reader = parser.getXMLReader();
				filter.setParent(reader);
				
				return faucetObject; // An InputSource really!
		    } catch (Exception e) {
		    	throw new XformationException("Unable to set up filter chain", e);
		    }
		}
	}

	@Override
	public void readFrom(DataInputStream in) throws IOException,
			IllegalAccessException, InstantiationException {
		this.id = in.readUTF();
		String sinkClassName = in.readUTF();
		Class<?> sinkClass;
		try {
			sinkClass = Class.forName(sinkClassName);
		} catch (ClassNotFoundException e) {
			throw new InstantiationException("Unable to load class: " + sinkClassName);
		}
		this.sink = (ISink) sinkClass.newInstance();
		this.sink.readFrom(in);
		this.hookedUpSink = true;
		this.sink.hookupFaucet(this);
		this.faucet = null;
		this.hookedUpFaucet = false;
		this.type = Type.File;
		this.faucetObject = null;
		this.styleSheetInputStream = null;
		this.styleSheetPath = in.readUTF();
		this.filter = null;
	}
	
	@Override
	public void writeTo(DataOutputStream out) throws IOException {
		if (type != Type.File)
			throw new XformationException(
					"Only XMLFilterPipelineItems of type Type.File can be streamed.");
		
		out.writeUTF(id);
		out.writeUTF(sink.getClass().getCanonicalName());
		sink.writeTo(out);
		out.writeUTF(styleSheetPath);
	}

	@Override
	public List<ISink> getTerminalSinks() {
		if (sink instanceof IFaucet)
			return ((IFaucet) sink).getTerminalSinks();
		else if (sink instanceof ISplitPipelineItem)
			return ((ISplitPipelineItem) sink).getTerminalSinks();
		else
			return Collections.singletonList(sink);
	}

}
