/*******************************************************************************
 * Copyright (c) 2009 Pafsanias Ftakas.
 *
 * This file is part of DistProc.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * DistProc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * Eclipse Public License v1.0 for more details.
 *
 * Contributors:
 *     Pafsanias Ftakas - initial API and implementation
 *******************************************************************************/
package com.koutra.dist.proc.xml;

import java.io.IOException;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Types;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.lang.StringEscapeUtils;
import org.apache.log4j.Logger;
import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.AttributesImpl;

import com.koutra.dist.proc.pipeline.ResultSetToXMLDemuxPipelineItem;
import com.koutra.dist.proc.util.ByteArrayUtil;

public class ResultSetSAXEventReader implements XMLReader {
	private static final Logger logger = Logger.getLogger(ResultSetSAXEventReader.class);
	
	private static final SimpleDateFormat dateSDF = new SimpleDateFormat("dd/MM/yyyy");
	private static final SimpleDateFormat timeSDF = new SimpleDateFormat("HH:mm:ss");
	private static final SimpleDateFormat dtSDF = new SimpleDateFormat("dd/MM/yyyy HH:mm:ss");
	
	private class DocumentLocator implements Locator {

		public int getColumnNumber() {
			return -1;
		}

		public int getLineNumber() {
			return -1;
		}

		public String getPublicId() {
			return null;
		}

		public String getSystemId() {
			return null;
		}
		
	}
	
	protected ContentHandler contentHandler;
	protected ResultSet resultSet;
	protected int columnNumber;
	protected Map<Integer, String> columnNames;
	protected Map<Integer, Integer> columnTypes;
	protected ResultSetToXMLDemuxPipelineItem demux;
	
	public ResultSetSAXEventReader(ResultSet resultSet,
			ResultSetToXMLDemuxPipelineItem demux) throws SQLException {
		this.contentHandler = null;
		this.resultSet = resultSet;
		this.demux = demux;
		
		ResultSetMetaData metadata = resultSet.getMetaData();
		columnNumber = metadata.getColumnCount();
		columnNames = new HashMap<Integer, String>();
		columnTypes = new HashMap<Integer, Integer>();
		for (int i = 1; i <= columnNumber; i++) {
			columnNames.put(i, metadata.getColumnName(i).toLowerCase());
			columnTypes.put(i, metadata.getColumnType(i));
		}
	}
	
	public ResultSetSAXEventReader(ResultSet resultSet) throws SQLException {
		this(resultSet, null);
	}
	
	protected void convertResultSet() throws SAXException, SQLException {
		if (demux == null) {
			contentHandler.startDocument();
			contentHandler.startElement("", "result", "result", new AttributesImpl());
		}
		int count = 1;
		while (resultSet.next()) {
			if (demux != null) {
				if (count != 1) {
					if (logger.isTraceEnabled())
						logger.trace("Starting pipeline switch");
					demux.switchPipeline();
					if (logger.isTraceEnabled())
						logger.trace("Ending pipeline switch");
				}
				
				contentHandler.startDocument();
				contentHandler.startElement("", "result", "result", new AttributesImpl());
			}
			AttributesImpl atts = new AttributesImpl();
			atts.addAttribute("", "count", "count", "CDATA", String.valueOf(count++));
			contentHandler.startElement("", "row", "row", atts);
			
			for (int i = 1; i <= columnNumber; i++) {
				int type = columnTypes.get(i);
				atts = new AttributesImpl();
				atts.addAttribute("", "type", "type", "CDATA", String.valueOf(type));
				String name = columnNames.get(i);
				contentHandler.startElement("", name, name, atts);
				String value = null;
				switch (type) {
				case Types.CHAR:
				case Types.VARCHAR:
				case Types.LONGVARCHAR:
				case Types.NCHAR:
				case Types.NVARCHAR:
				case Types.LONGNVARCHAR:
					value = StringEscapeUtils.escapeXml(resultSet.getString(i));
					break;
				case Types.BIT:
				case Types.BOOLEAN:
					value = resultSet.getBoolean(i) ? "true" : "false";
					break;
				case Types.TINYINT:
					value = String.valueOf(resultSet.getByte(i));
					break;
				case Types.SMALLINT:
					value = String.valueOf(resultSet.getShort(i));
					break;
				case Types.INTEGER:
					value = String.valueOf(resultSet.getInt(i));
					break;
				case Types.BIGINT:
					value = String.valueOf(resultSet.getLong(i));
					break;
				case Types.REAL:
					value = String.valueOf(resultSet.getFloat(i));
					break;
				case Types.FLOAT:
				case Types.DOUBLE:
					value = String.valueOf(resultSet.getDouble(i));
					break;
				case Types.NUMERIC:
				case Types.DECIMAL:
					value = String.valueOf(resultSet.getBigDecimal(i));
					break;
				case Types.BINARY:
				case Types.VARBINARY:
				case Types.LONGVARBINARY:
					byte[] buffer = resultSet.getBytes(i);
					value = ByteArrayUtil.encodeInHex(buffer, 0, buffer.length);
					break;
				case Types.DATE:
					value = dateSDF.format(resultSet.getDate(i));
					break;
				case Types.TIME:
					value = timeSDF.format(resultSet.getTime(i));
					break;
				case Types.TIMESTAMP:
					value = dtSDF.format(resultSet.getTimestamp(i));
					break;
				default:
					value = "UNRECOGNIZED TYPE: " + type;
					break;
				}
				contentHandler.characters(value.toCharArray(), 0, value.length());
				contentHandler.endElement("", name, name);
			}
			
			contentHandler.endElement("", "row", "row");
			
			if (demux != null) {
				contentHandler.endElement("", "result", "result");
				contentHandler.endDocument();
			}
		}
		if (demux == null) {
			contentHandler.endElement("", "result", "result");
			contentHandler.endDocument();
		} else {
			if (count > 1) {
				// We have gone round the big while loop at least once.
				// We therefore have SAX events we need to flush down the pipeline.
				if (logger.isTraceEnabled())
					logger.trace("Starting pipeline switch");
				demux.switchPipeline();
				if (logger.isTraceEnabled())
					logger.trace("Ending pipeline switch");
			}
			demux.signalTermination();
		}
	}

	public void parse(InputSource input) throws IOException, SAXException {
		contentHandler.setDocumentLocator(new DocumentLocator());
		try {
			convertResultSet();
		} catch (SQLException e) {
			throw new SAXException("Unable to convert result set", e);
		}
	}

	public void parse(String systemId) throws IOException, SAXException {
		contentHandler.setDocumentLocator(new DocumentLocator());
		try {
			convertResultSet();
		} catch (SQLException e) {
			throw new SAXException("Unable to convert result set", e);
		}
	}

	public ContentHandler getContentHandler() {
		return contentHandler;
	}

	public void setContentHandler(ContentHandler handler) {
		contentHandler = handler;
		if (logger.isTraceEnabled())
			logger.trace("Switched content handler to: " + contentHandler);
	}

	//////////////////////////////////////////////////////////////////////////////////////
	// The rest of the implementation is not used.
	//////////////////////////////////////////////////////////////////////////////////////
	
	public DTDHandler getDTDHandler() {
		return null;
	}

	public EntityResolver getEntityResolver() {
		return null;
	}

	public ErrorHandler getErrorHandler() {
		return null;
	}

	public boolean getFeature(String name) throws SAXNotRecognizedException,
			SAXNotSupportedException {
		return false;
	}

	public Object getProperty(String name) throws SAXNotRecognizedException,
			SAXNotSupportedException {
		return null;
	}

	public void setDTDHandler(DTDHandler handler) {
	}

	public void setEntityResolver(EntityResolver resolver) {
	}

	public void setErrorHandler(ErrorHandler handler) {
	}

	public void setFeature(String name, boolean value)
			throws SAXNotRecognizedException, SAXNotSupportedException {
	}

	public void setProperty(String name, Object value)
			throws SAXNotRecognizedException, SAXNotSupportedException {
	}

}
